@@ -3,11 +3,11 @@
  * The Kronolith_Driver_sql:: class implements the Kronolith_Driver
  * API for a SQL backend.
  *
- * $Horde: kronolith/lib/Driver/sql.php,v 1.31.2.9 2003/04/04 19:24:01 chuck Exp $
+ * $Horde: kronolith/lib/Driver/sql.php,v 1.31.2.13 2004/03/05 19:26:04 jan Exp $
  *
  * @author  Luc Saillard <luc.saillard@fr.alcove.com>
  * @author  Chuck Hagenbuch <chuck@horde.org>
- * @version $Revision: 1.31.2.9 $
+ * @version $Revision: 1.31.2.13 $
  * @since   Kronolith 0.3
  * @package kronolith
  */
@@ -26,6 +26,13 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
      */
     var $_connected = false;
 
+    /**
+     * Cache events as we fetch them to avoid fetching the same event
+     * from the db twice.
+     * @var array $_cache
+     */
+    var $_cache = array();
+
     function open($calendar)
     {
         $this->_calendar = $calendar;
@@ -37,21 +44,21 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
         $allevents = $this->listEvents($date, $date, true);
         $events = array();
 
-        foreach ($allevents as $eventid) {
-            $event = $this->getEventObject($eventid);
+        foreach ($allevents as $eventId) {
+            $event = &$this->getEventObject($eventId);
 
             if ($event->getRecurType() == KRONOLITH_RECUR_NONE) {
                 $diff = Kronolith::dateDiff($date,
                                             Kronolith::timestampToObject($event->startTimestamp));
                 if ($diff->sec < $event->getAlarm() * 60) {
-                    $events[] = $eventid;
+                    $events[] = $eventId;
                 }
             } else {
-                $next = $this->nextRecurrence($eventid, $date);
+                $next = $this->nextRecurrence($eventId, $date);
                 if ($next) {
                     $diff = Kronolith::dateDiff($date, $next);
                     if ($diff->sec < $event->getAlarm() * 60) {
-                        $events[] = $eventid;
+                        $events[] = $eventId;
                     }
                 }
             }
@@ -62,8 +69,6 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
 
     function listEvents($startDate = null, $endDate = null, $hasAlarm = false)
     {
-        $events = array();
-
         if (!isset($endDate)) {
             $endDate = Kronolith::dateObject(array('mday' => 31, 'month' => 12, 'year' => 9999));
         } else {
@@ -80,22 +85,26 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
             $stime = sprintf('%04d-%02d-%02d 00:00:00', $startDate->year, $startDate->month, $startDate->mday);
         }
 
-        $q = 'SELECT DISTINCT e.event_id, e.event_recurtype FROM ' . $this->_params['table'] . ' e' .
-             ' WHERE e.calendar_id = ' . $this->_db->quote($this->_calendar) . ' AND ((';
+        $q = 'SELECT DISTINCT event_id, event_description, event_location,' .
+            ' event_keywords, event_title, event_category,' .
+            ' event_recurtype, event_recurenddate, event_recurinterval,' .
+            ' event_recurdays, event_start, event_end, event_alarm,' .
+            ' event_modified, event_exceptions, event_creator_id FROM ' . $this->_params['table'] .
+            ' WHERE calendar_id = ' . $this->_db->quote($this->_calendar) . ' AND ((';
 
         if ($hasAlarm) {
-            $q .= 'e.event_alarm > 0)) AND ((';
+            $q .= 'event_alarm > 0)) AND ((';
         }
 
         if (isset($stime)) {
-            $q .= 'e.event_end > ' . $this->_db->quote($stime) . ' AND ';
+            $q .= 'event_end > ' . $this->_db->quote($stime) . ' AND ';
         }
-        $q .= 'e.event_start < ' . $this->_db->quote($etime) . ') OR (';
+        $q .= 'event_start < ' . $this->_db->quote($etime) . ') OR (';
         if (isset($stime)) {
-            $q .= 'e.event_recurenddate >= ' . $this->_db->quote($stime) . ' AND ';
+            $q .= 'event_recurenddate >= ' . $this->_db->quote($stime) . ' AND ';
         }
-        $q .= 'e.event_start <= ' . $this->_db->quote($etime) .
-              ' AND e.event_recurtype != ' . KRONOLITH_RECUR_NONE . '))';
+        $q .= 'event_start <= ' . $this->_db->quote($etime) .
+              ' AND event_recurtype != ' . KRONOLITH_RECUR_NONE . '))';
 
         /* Log the query at a DEBUG log level. */
         Horde::logMessage(sprintf('SQL event list by %s: table = %s; query = "%s"',
@@ -105,9 +114,15 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
         /* Run the query. */
         $qr = $this->_db->query($q);
 
+        $events = array();
         if (!PEAR::isError($qr)) {
             $row = $qr->fetchRow(DB_FETCHMODE_ASSOC);
             while ($row && !PEAR::isError($row)) {
+                // We have all the information we need to create an
+                // event object for this event, so go ahead and cache
+                // it.
+                $this->_cache[$this->_calendar][$row['event_id']] = &new Kronolith_Event_sql($this, $row);
+
                 if ($row['event_recurtype'] == KRONOLITH_RECUR_NONE) {
                     $events[] = $row['event_id'];
                 } else {
@@ -124,25 +139,29 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
         return $events;
     }
 
-    function getEventObject($eventID = null)
+    function &getEventObject($eventId = null)
     {
-        if (!is_null($eventID)) {
+        if (!is_null($eventId)) {
+            if (isset($this->_cache[$this->_calendar][$eventId])) {
+                return $this->_cache[$this->_calendar][$eventId];
+            }
+
             $event = &$this->_db->getRow('SELECT event_id, event_description, event_location,' .
                                          ' event_keywords, event_title, event_category,' .
                                          ' event_recurtype, event_recurenddate, event_recurinterval,' .
                                          ' event_recurdays, event_start, event_end, event_alarm,' .
                                          ' event_modified, event_exceptions' .
                                          ' FROM ' . $this->_params['table'] .
-                                         ' WHERE event_id = ' . (int)$eventID .
+                                         ' WHERE event_id = ' . (int)$eventId .
                                          ' AND calendar_id = ' . $this->_db->quote($this->_calendar),
                                          DB_FETCHMODE_ASSOC);
-
             if (PEAR::isError($event)) {
                 return $event;
             }
 
             if ($event) {
-                return new Kronolith_Event_sql($this, $event);
+                $this->_cache[$this->_calendar][$eventId] = &new Kronolith_Event_sql($this, $event);
+                return $this->_cache[$this->_calendar][$eventId];
             } else {
                 return false;
             }
@@ -205,38 +224,30 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
         }
     }
 
-    function nextRecurrence($eventID, $afterDate, $weekstart = KRONOLITH_SUNDAY)
+    function nextRecurrence($eventId, $afterDate)
     {
-        $daysInMonth = Kronolith::daysInMonth($afterDate['month'], $afterDate['year']);
-        while ($afterDate['mday'] > $daysInMonth) {
-            $afterDate['mday'] -= $daysInMonth;
-            $afterDate['month']++;
-            if ($afterDate['month'] > 12) {
-                $afterDate['year']++;
-                $afterDate['month'] -= 12;
-            }
-        }
+        $afterDate = Kronolith::dateObject($afterDate);
+        $afterDate = Kronolith::correctDate($afterDate);
 
-        $event = $this->getEventObject($eventID);
+        $event = &$this->getEventObject($eventId);
         if (PEAR::isError($event)) {
             return $event;
         }
 
         $afterDate = Kronolith::dateObject($afterDate);
 
-        if (Kronolith::compareDates($event->start, $afterDate) > 0) {
+        if (Kronolith::compareDates($event->start, $afterDate) >=0) {
             return $event->start;
         }
 
         $event->recurEnd->hour = 23;
-        $event->recurEnd->min  = 59;
-        $event->recurEnd->sec  = 59;
+        $event->recurEnd->min = 59;
+        $event->recurEnd->sec = 59;
 
         switch ($event->getRecurType()) {
-
         case KRONOLITH_RECUR_DAILY:
             $diff = Kronolith::dateDiff($event->start, $afterDate);
-            $recur = (ceil($diff->mday / $event->recurInterval)) * $event->recurInterval;
+            $recur = ceil($diff->mday / $event->recurInterval) * $event->recurInterval;
             $next = $event->start;
             list($next->mday, $next->month, $next->year) = explode('/', Date_Calc::daysToDate(Date_Calc::dateToDays($next->mday, $next->month, $next->year) + $recur, '%e/%m/%Y'));
             if (Kronolith::compareDates($next, $event->recurEnd) <= 0 &&
@@ -264,7 +275,7 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
             }
             if (Kronolith::compareDates($next, $event->recurEnd) <= 0) {
                 if (Kronolith::compareDates($next, $after_week_end) >= 0) {
-                    return $this->nextRecurrence($eventID, $after_week_end);
+                    return $this->nextRecurrence($eventId, $after_week_end);
                 }
                 while (!$event->recurOnDay((int)pow(2, (int)Date_Calc::dayOfWeek($next->mday, $next->month, $next->year))) && Kronolith::compareDates($next, $after_week_end) < 0) {
                     $next->mday++;
@@ -272,7 +283,7 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
                 }
                 if (Kronolith::compareDates($next, $event->recurEnd) <= 0) {
                     if (Kronolith::compareDates($next, $after_week_end) >= 0) {
-                        return $this->nextRecurrence($eventID, $after_week_end);
+                        return $this->nextRecurrence($eventId, $after_week_end);
                     } else {
                         return Kronolith::dateObject($next);
                     }
@@ -293,43 +304,75 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
             break;
 
         case KRONOLITH_RECUR_WEEK_OF_MONTH:
-            $week = ceil($event->start->mday / 7);
-            $wday = date('w', Kronolith::objectToTimestamp($event->start));
+            // Start with the start date of the event.
+            $estart = Kronolith::dateObject($event->start);
 
-            $diff = Kronolith::dateDiff($event->start, $afterDate);
-            $recur = $diff->month + $diff->month % $event->recurInterval;
-            $next = $event->start;
-            $next->month += $recur;
-            $next = Kronolith::correctDate($next);
-            $next->mday = $week*7;
-            $next_wday = date('w', Kronolith::objectToTimestamp($next));
-            if ($next_wday > $wday) {
-                $next->mday -= ($next_wday - $wday);
-            } elseif ($next_wday < $wday) {
-                $next->mday -= ($next_wday + (7 - $wday));
-            }
+            // What day of the week, and week of the month, do we
+            // recur on?
+            $nth = ceil($event->start->mday / 7);
+            $weekday = Kronolith::dayOfWeek($estart->year, $estart->month, $estart->mday);
 
-            // Correct the data and return it if w/in the requested range
-            $next = Kronolith::correctDate($next);
-            if (Kronolith::compareDates($next, $event->recurEnd) <= 0 &&
-                Kronolith::compareDates($next, $afterDate) >= 0) {
-                return $next;
-            }
-            break;
+            // Adjust $estart to be the first candidate.
+            $offset = ($afterDate->month - $estart->month) + ($afterDate->year - $estart->year) * 12;
+            $offset = floor(($offset + $event->recurInterval - 1) / $event->recurInterval) * $event->recurInterval;
+
+            // Adjust our working date until it's after $afterDate.
+            $estart->month += $offset - $event->recurInterval;
+            do {
+                $estart->month += $event->recurInterval;
+                $estart = Kronolith::correctDate($estart);
+
+                $next = Kronolith::dateObject($estart);
+                $next = Kronolith::setNthWeekday($next, $weekday, $nth);
+
+                if (Kronolith::compareDates($next, $afterDate) < 0) {
+                    // We haven't made it past $afterDate yet, try
+                    // again.
+                    continue;
+                }
+                if (Kronolith::compareDates($next, $event->recurEnd) > 0) {
+                    // We've gone past the end of recurrence; we can
+                    // give up now.
+                    return false;
+                }
+
+                // We have a candidate to return.
+                break;
+            } while (true);
+
+            return $next;
 
         case KRONOLITH_RECUR_YEARLY:
-            $diff = Kronolith::dateDiff($event->start, $afterDate);
-            $recur = $diff->year + $diff->year % $event->recurInterval;
-            $next = $event->start;
-            $next->year += $recur;
-            if (Kronolith::compareDates($next, $event->recurEnd) <= 0 &&
-                Kronolith::compareDates($next, $afterDate) >= 0) {
-                return $next;
-            }
-            break;
+            // Start with the start date of the event.
+            $estart = Kronolith::dateObject($event->start);
 
+            // We probably need a seperate case here for February 29th
+            // and leap years, but until we're absolutely sure it's a
+            // bug, we'll leave it out.
+            if ($afterDate->month > $estart->month ||
+                ($afterDate->month == $estart->month && $afterDate->mday > $estart->mday)) {
+                $afterDate->year++;
+                $afterDate->month = $estart->month;
+                $afterDate->mday = $estart->mday;
+            }
+
+            // Adjust $estart to be the first candidate.
+            $offset = $afterDate->year - $estart->year;
+            if ($offset > 0) {
+                $offset = (($offset + $event->recurInterval - 1) / $event->recurInterval) * $event->recurInterval;
+                $estart->year += $offset;
+            }
+
+            // We've gone past the end of recurrence; give up.
+            if (Kronolith::compareDates($event->recurEnd, $estart) < 0) {
+                return false;
+            }
+
+            return $estart;
         }
 
+        // We didn't find anything, the recurType was bad, or
+        // something else went wrong - return false.
         return false;
     }
 
@@ -359,16 +402,16 @@ class Kronolith_Driver_sql extends Kronolith_Driver {
     /**
      * Delete an event.
      *
-     * @param string $eventID The id of the event to delete.
+     * @param string $eventId The id of the event to delete.
      *
      * @return mixed  True or a PEAR_Error on failure.
      */
-    function deleteEvent($eventID)
+    function deleteEvent($eventId)
     {
-        $eventID = (int)$eventID;
+        $eventId = (int)$eventId;
         $query = sprintf('DELETE FROM %s WHERE event_id = %s AND calendar_id = %s',
                          $this->_params['table'],
-                         $this->_db->quote($eventID),
+                         $this->_db->quote($eventId),
                          $this->_db->quote($this->_calendar));
 
         /* Log the query at a DEBUG log level. */
@@ -465,8 +508,8 @@ class Kronolith_Event_sql extends Kronolith_Event {
     {
         $driver = &$this->getDriver();
 
-        $this->start = new stdClass();
-        $this->end = new stdClass();
+        $this->start = &new stdClass();
+        $this->end = &new stdClass();
         list($this->start->year, $this->start->month, $this->start->mday, $this->start->hour, $this->start->min, $this->start->sec) = sscanf($SQLEvent['event_start'], '%04d-%02d-%02d %02d:%02d:%02d');
         list($this->end->year, $this->end->month, $this->end->mday, $this->end->hour, $this->end->min, $this->end->sec) = sscanf($SQLEvent['event_end'], '%04d-%02d-%02d %02d:%02d:%02d');
 
@@ -476,7 +519,7 @@ class Kronolith_Event_sql extends Kronolith_Event {
         $this->durMin = ($this->endTimestamp - $this->startTimestamp) / 60;
 
         if (isset($SQLEvent['event_recurenddate'])) {
-            $this->recurEnd = new stdClass();
+            $this->recurEnd = &new stdClass();
             list($this->recurEnd->year, $this->recurEnd->month, $this->recurEnd->mday, $this->recurEnd->hour, $this->recurEnd->min, $this->recurEnd->sec) = sscanf($SQLEvent['event_recurenddate'], '%04d-%02d-%02d %02d:%02d:%02d');
             $this->recurEndTimestamp = @mktime($this->recurEnd->hour, $this->recurEnd->min, $this->recurEnd->sec, $this->recurEnd->month, $this->recurEnd->mday, $this->recurEnd->year);
         }
